#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['qval']

package_data = \
{'': ['*']}

setup(name='qval',
      version='0.1.5',
      description='This module provides convenient API for verifying query parameters.',
      author='George Usan-Podgornov',
      author_email='',
      url='https://github.com/OptimalStrategy/Qval',
      packages=packages,
      package_data=package_data,
     )
